
/*----------------------------------------------------------------------------+
 |                                                                            |
 |                RF Auxiliary Input Support for HEYU                         |
 |              Copyright 2006, 2007 Charles W. Sullivan                      |
 |                      All Rights Reserved                                   |
 |                                                                            |
 |                                                                            |
 | This software is licensed free of charge for non-commercial distribution   |
 | and for personal and internal business use only.  Inclusion of this        |
 | software or any part thereof in a commercial product is prohibited         |
 | without the prior written permission of the author.  You may copy, use,    |
 | and distribute this software subject to the following restrictions:        |
 |                                                                            |
 |  1)	You may not charge money for it.                                      |
 |  2)	You may not remove or alter this license, copyright notice, or the    |
 |      included disclaimers.                                                 |
 |  3)	You may not claim you wrote it.                                       |
 |  4)	If you make improvements (or other changes), you are requested        |
 |	to send them to the Heyu maintainer so there's a focal point for      |
 |      distributing improved versions.                                       |
 |                                                                            |
 | As used herein, HEYU is a trademark of Daniel B. Suthers.                  | 
 | X10, CM11A, and ActiveHome are trademarks of X-10 (USA) Inc.               |
 | The author is not affiliated with either entity.                           |
 |                                                                            |
 | Charles W. Sullivan                                                        |
 | Co-author and Maintainer                                                   |
 | Greensboro, North Carolina                                                 |
 | Email ID: cwsulliv01                                                       |
 | Email domain: -at- heyu -dot- org                                          |
 |                                                                            |
 | Disclaimers:                                                               |
 | THERE IS NO ASSURANCE THAT THIS SOFTWARE IS FREE OF DEFECTS AND IT MUST    |
 | NOT BE USED IN ANY SITUATION WHERE THERE IS ANY CHANCE THAT ITS            |
 | PERFORMANCE OR FAILURE TO PERFORM AS EXPECTED COULD RESULT IN LOSS OF      |
 | LIFE, INJURY TO PERSONS OR PROPERTY, FINANCIAL LOSS, OR LEGAL LIABILITY.   |
 |                                                                            |
 | TO THE EXTENT ALLOWED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED "AS IS",|
 | WITH NO EXPRESS OR IMPLIED WARRANTY, INCLUDING, BUT NOT LIMITED TO, THE    |
 | IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.|
 |                                                                            |
 | IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL THE AUTHOR BE LIABLE    |
 | FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   |
 | DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE EVEN IF   |
 | THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.            |
 |                                                                            |
 +----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>
#include "x10.h"
#include "process.h"

#ifdef pid_t
#define PID_T pid_t
#else
#define PID_T long
#endif

#define W800_LEN 4
#define PACKET_LEN  4
#define BSIZE (100 * PACKET_LEN)


#define HC_MASK      0xF000
#define FUNC_MASK    0x00B8
#define UNIT_MASK    0x0458
#define NOSW_MASK    0x0003
#define OFF_MASK     0x0020
#define OLD_MASK     0x0800
#define NEW_MASK     0x0040
#define STD_MASK     (HC_MASK | FUNC_MASK | UNIT_MASK)
#define NOTX10_MASK  ~(HC_MASK|FUNC_MASK|UNIT_MASK|OLD_MASK)

extern int tty_aux;
extern int sptty;
extern int tty;
extern int verbose;
extern int heyu_parent;
extern int sxread();
extern int i_am_aux, i_am_relay;
extern PID_T lockpid();
extern int lock_for_write(), munlock();


extern CONFIG config, *configp;



static unsigned char rf2hcode[16] = {
   0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15
};

#if 0
/* RF Codes for X10 Housecodes (A-P) */
static unsigned int hc2rfcode[16] = {
   0x6000, 0x7000, 0x4000, 0x5000,
   0x8000, 0x9000, 0xA000, 0xB000,
   0xE000, 0xF000, 0xC000, 0xD000,
   0x0000, 0x1000, 0x2000, 0x3000,
};   

/* Housecode for RF code (upper nybble) */
static char *rf2hc = "MNOPCDABEFGHKLIJ";
#endif

/* RF Codes for X10 _encoded_ unit codes */
static unsigned int rf_unit_code[16] = {
   0x0440, 0x0040, 0x0008, 0x0408,
   0x0448, 0x0048, 0x0000, 0x0400,
   0x0450, 0x0050, 0x0018, 0x0418,
   0x0458, 0x0058, 0x0010, 0x0410,
};      

/* RF Function Codes:               */
/* AllOff, LightsOn, On, Off,       */
/* Dim,    Bright, LightsOff        */
static unsigned int rf_func_code[] = {
   0x0080, 0x0090, 0x0000, 0x0020,
   0x0098, 0x0088, 0x00a0,
};


char *funclist[7] =
   {"alloff", "lightson", "on", "off", "dim", "bright", "lightsoff"};

/*-------------------------------------------------------------+
 | Check list of valid 4-byte codes other than Standard,       |
 | Entertainment, and Security codes, e.g., Pan/Tilt           |
 | Return type code if valid, otherwise RF_NOISEW800.          |
 | (*** Future ***)                                            |
 +-------------------------------------------------------------*/
int is_valid_w800_code( long code )
{
      
   return RF_NOISEW800;
}


/*-------------------------------------------------------------+
 | Translate Standard X10 RF code into hcode, ucode, and fcode |
 +-------------------------------------------------------------*/
int rf_parms (int rfword, unsigned char *hcode, unsigned char *ucode, unsigned char *fcode)
{
   unsigned int  rem;
   unsigned char is_unit;
   int           j;

   /* Old X10 wireless switch code to new code */
   if ( rfword & OLD_MASK ) {
      rfword = (rfword & ~OLD_MASK) | NEW_MASK;
   }

   /* Standardize */
   rfword &= STD_MASK;

   *hcode = rf2hcode[(rfword & HC_MASK) >> 12];

   rem = rfword & ~HC_MASK & ~OFF_MASK;

   is_unit = 0;
   for ( j = 0; j < 16; j++ ) {
      if ( rem == rf_unit_code[j] ) {
         *ucode = j;
         is_unit = 1;
         *fcode = (rfword & OFF_MASK) ? 3 : 2;
         break;
      }
   }

   if ( !is_unit ) {
      *ucode = 0xff;
      rem = rfword & ~HC_MASK;
      for ( j = 0; j < 7; j++ ) {
         if ( rem == rf_func_code[j] ) {
            *fcode = j;
            break;
         }
      }
   }

   return 0;
}

      
/*-------------------------------------------------------------+
 | Process RF data of type RF_STD (Standard X10 RF remotes)  |
 | and send to spoolfile and interface per config.             |
 +-------------------------------------------------------------*/
int proc_type_std( unsigned char hcode, unsigned char ucode,
                                            unsigned char fcode )
{
   unsigned char cmdbuf[16];
   char writefilename[PATH_LEN + 1];

   if ( (configp->rf_funcmask & (1 << fcode)) == 0 ||
         configp->transceive[hcode] == 0 ) {
      return 0;
   }

   sprintf(writefilename, "%s%s", WRITEFILE, configp->suffix);


   if ( lock_for_write() < 0 ) {
      syslog(LOG_ERR, "aux unable to lock writefile\n");
      return 1;
   }

   switch ( fcode ) {
      case 2 : /* On */
      case 3 : /* Off */
         if ( (configp->transceive[hcode] & (1 << ucode)) == 0 )
            break;
         send_address(hcode, (1 << ucode), NO_SYNC);
         cmdbuf[0] = 0x06;
         cmdbuf[1] = (hcode << 4) | fcode;
         send_command(cmdbuf, 2, 0, NO_SYNC);
         break;
      case 4 : /* Dim */
      case 5 : /* Bright */
         cmdbuf[0] = 0x06 | (configp->trans_dim << 3);
         cmdbuf[1] = (hcode << 4) | fcode;
         send_command(cmdbuf, 2, 0, NO_SYNC);
         break;
      case 0 : /* AllOff */
      case 1 : /* LightsOn */
      case 6 : /* LightsOff */
      default :
         cmdbuf[0] = 0x06;
         cmdbuf[1] = (hcode << 4) | fcode;
         send_command(cmdbuf, 2, 0, NO_SYNC);
         break;
   }

   munlock(writefilename);


   return 0;
}

#if 0
/*-------------------------------------------------------------+
 | Process RF data of types RF_SEC (Security RF remotes) and   |
 | RF_ENT (Entertainment RF remotes)                           |
 +-------------------------------------------------------------*/
int proc_type_sec_ent ( unsigned char vtype, unsigned char ident,
                                            unsigned char vdata )
{
   ALIAS         *aliasp;
   unsigned char hcode, ucode;
   unsigned int  bitmap;
   int           j, k, found;
   char writefilename[PATH_LEN + 1];

   if ( !(aliasp = configp->aliasp) ) {
      send_virtual_aux_data(0, vdata, vtype, ident, 0, 0);
      return 0;
   }

   hcode = ucode = 0;
   found = 0;
   j = 0;
   while ( !found && aliasp[j].line_no > 0 ) {
      if ( aliasp[j].vtype == vtype && 
           (bitmap = aliasp[j].unitbmap) > 0 &&
           (ucode = single_bmap_unit(bitmap)) != 0xff ) {
         for ( k = 0; k < aliasp[j].nident; k++ ) {
            if ( aliasp[j].ident[k] == ident ) {
               hcode = hc2code(aliasp[j].housecode);
               found = 1;
               break;
            }
         }
      }
      j++;
   }

   sprintf(writefilename, "%s%s", WRITEFILE, configp->suffix);

   if ( lock_for_write() < 0 ) {
      syslog(LOG_ERR, "aux unable to lock writefile\n");
      return 1;
   }

   if ( found )
      send_virtual_aux_data((hcode << 4) | ucode, vdata, vtype | ADDR_STD, ident, 0, 0);
   else
      send_virtual_aux_data(0, vdata, vtype, ident, 0, 0);

   munlock(writefilename);

   return 0;
}
#endif

/*-------------------------------------------------------------+
 | Alert to an RF Flood                                        |
 +-------------------------------------------------------------*/
int send_rf_flood_message( unsigned char vflags )
{
   char writefilename[PATH_LEN + 1];

   sprintf(writefilename, "%s%s", WRITEFILE, configp->suffix);

   if ( lock_for_write() < 0 ) {
      syslog(LOG_ERR, "aux unable to lock writefile\n");
      return 1;
   }


#ifdef USE_ST_COMMAND
   send_x10state_command(ST_RFFLOOD, vflags);
#else
   send_virtual_aux_data(0, vflags, RF_FLOOD, 0, 0, 0);
#endif

   munlock(writefilename);

   return 0;
}

int write_restart_error ( char *restartfile )
{
   FILE *fp;
  
   if ( (fp = fopen(restartfile, "w")) == NULL ) {
      syslog(LOG_ERR, "Cannot fopen restartfile\n");
      return 1;
   }

   fprintf(fp, "%s\n", error_message());

   fclose(fp);

   return 0;
}

#define ABSMAX_IN_A_ROW  1000000
   
enum {MinCount, RepCount, MaxCount, FloodRep};

/*-------------------------------------------------------------+
 | Process data read from the W800RF32 on the TTY_AUX serial   |
 | port and send to spoolfile and/or interface per config.     |
 +-------------------------------------------------------------*/
int aux_w800 ( void )
{
   unsigned char type;
   unsigned char hcode, ucode, fcode, rfflood;
   unsigned char buff[8];
   unsigned long in_a_row, max_in_a_row;
   long          rfword, last_word, this_word;
   int           count = 0;
   int           nread = 0;
   int           is_idle = 0;
   int           is_err = 0;
   int           new_burst = 0;
   char          restartfile[PATH_LEN + 1];
   struct stat   statbuf;
   extern void aux_local_setup(void);

   rfword = last_word = -1;
   type = 0;
   nread = 0;
   is_idle = 0;
   in_a_row = 0;
   rfflood = 0;
   max_in_a_row = config.aux_repcounts[MaxCount];
   
   strncpy2(restartfile, pathspec(RESTART_AUX_FILE), PATH_LEN);
   unlink(restartfile);

   openlog( "heyu_aux", 0, LOG_DAEMON);

   while ( 1 )  {
      /* Check if restart needed */
      if ( is_idle && stat(restartfile, &statbuf) == 0 && is_err == 0 ) {
         if ( reread_config() != 0 ) {
            is_err = 1;
            syslog(LOG_ERR, "aux reconfiguration failed!\n");
            write_restart_error(restartfile);
         }
         else {
            syslog(LOG_ERR, "aux reconfiguration-\n");
            unlink(restartfile);
            return 0;
         }
      }

      while ( nread < 4 && sxread(tty_aux, buff + nread, 1, 1) ) {
         nread++;
      }

      if ( nread < 4 ) {
         is_idle = 1;
         nread = 0;
         count = 0;
         in_a_row = 0;
         max_in_a_row = config.aux_repcounts[MaxCount];
         if ( max_in_a_row > 0 && rfflood > 0 )
            send_rf_flood_message(0); /* End of flood */
         rfflood = 0;
         continue;
      }
      is_idle = 0;

      /* Check for a continuous flood of RF inputs without a break */
      if ( max_in_a_row > 0 && ++in_a_row == max_in_a_row ) {
         rfflood = 1;
         send_rf_flood_message(SEC_FLOOD);
         max_in_a_row *= config.aux_repcounts[FloodRep];
         if ( max_in_a_row > ABSMAX_IN_A_ROW ) {
            max_in_a_row = ABSMAX_IN_A_ROW;
            in_a_row = 0;
         }
         continue;
      }


      if ( (buff[2] ^ buff[3]) != 0xffu ) {
         this_word = (buff[0] << 24) | (buff[1] << 16) | (buff[2] << 8) | buff[3];
         if ( (type = is_valid_w800_code(this_word)) == RF_NOISEW800 ) {
            /* Noise */
            if ( config.disp_raw_rf & DISPMODE_RF_NOISE )
               send_virtual_aux_data(0, buff[0], RF_NOISEW800, buff[1], buff[2], buff[3]);
            nread = 0;
            continue;
         }
      }
      else if ( (buff[0] ^ buff[1]) == 0xffu ) {
         /* X10 Standard RF (type 0) or Entertainment RF (type 1) */
         type = (buff[0] & 0x02) ? RF_ENT : RF_STD;
         this_word = (buff[0] << 8 | buff[2]);
      }
      else if ( (buff[0] ^ buff[1]) == 0x0fu ) {
         /* X10 Security RF */
         type = RF_SEC;
         this_word = (buff[0] << 8 | buff[2]);
      }
      else {
         this_word = (buff[0] << 24) | (buff[1] << 16) | (buff[2] << 8) | buff[3];
         if ( (type = is_valid_w800_code(this_word)) == RF_NOISEW800 ) {
            /* Noise */
            if ( config.disp_raw_rf & DISPMODE_RF_NOISE )
               send_virtual_aux_data(0, buff[0], RF_NOISEW800, buff[1], buff[2], buff[3]);
            nread = 0;
            continue;
         }
      }

      if ( (config.disp_raw_rf & DISPMODE_RF_NORMAL) && !new_burst ) {
         send_virtual_aux_data(0, buff[0], RF_RAWW800, buff[1], buff[2], buff[3]);
      }
      new_burst = 0;

      if ( count == 0 ) {
         rfword = this_word;
         count++;
         nread = 0;
         if ( type == RF_STD ) {
            rf_parms((int)rfword, &hcode, &ucode, &fcode);
            if ( count == config.aux_repcounts[MinCount] ) {
               proc_type_std(hcode, ucode, fcode);
            }
         }
         else if ( type == RF_SEC || type == RF_ENT ) {
            if ( count == config.aux_repcounts[MinCount] ) {
               send_virtual_aux_data(0, buff[2], type, buff[0], 0, 0);
            }
         }
         else if ( count == config.aux_repcounts[MinCount] ) {
            send_virtual_aux_data(0, buff[0], type, buff[1], buff[2], buff[3]);
         }
      }
      else if ( this_word == rfword ) {
         /* Repeat of previous burst */
         count++;
         nread = 0;
         if ( config.aux_repcounts[RepCount] > config.aux_repcounts[MinCount] &&
              (count % config.aux_repcounts[RepCount]) == config.aux_repcounts[MinCount] &&
              (config.aux_repcounts[MaxCount] == 0 ||
               count < config.aux_repcounts[MaxCount]) ) {
            if ( type == RF_STD ) {
               rf_parms((int)rfword, &hcode, &ucode, &fcode);
               proc_type_std(hcode, ucode, fcode);
            }
            else if ( type == RF_SEC || type == RF_ENT ) {
               send_virtual_aux_data(0, buff[2], type, buff[0], 0, 0);
            }
            else {
               send_virtual_aux_data(0, buff[0], type, buff[1], buff[2], buff[3]);
            }
         }
      }
      else {
         /* New burst, save buffer */
         new_burst = 1;
         count = 0;
      }
   }
   return 0;
}

/*-------------------------------------------------------------+
 | Process data read from the MR26A on the TTY_AUX serial      |
 | port and send to spoolfile and/or interface per config.     |
 +-------------------------------------------------------------*/
int aux_mr26a ( void )
{
   unsigned char type;
   unsigned char hcode, ucode, fcode, rfflood;
   unsigned char buff[8];
   unsigned long in_a_row, max_in_a_row;
   long          rfword, last_word, this_word;
   int           k;
   int           count = 0;
   int           nread = 0;
   int           is_idle = 0;
   int           is_err = 0;
   int           new_burst = 0;
   char          restartfile[PATH_LEN + 1];
   struct stat   statbuf;
   extern void aux_local_setup(void);

   rfword = last_word = -1;
   type = 0;
   nread = 0;
   is_idle = 0;
   in_a_row = 0;
   rfflood = 0;
   max_in_a_row = config.aux_repcounts[MaxCount];

   strncpy2(restartfile, pathspec(RESTART_AUX_FILE), PATH_LEN);
   unlink(restartfile);
   openlog( "heyu_aux", 0, LOG_DAEMON);

   while ( 1 )  {
      /* Check if restart needed */
      if ( is_idle && stat(restartfile, &statbuf) == 0 && is_err == 0 ) {
         if ( reread_config() != 0 ) {
            is_err = 1;
            syslog(LOG_ERR, "aux reconfiguration failed!\n");
            write_restart_error(restartfile);
         }
         else {
            syslog(LOG_ERR, "aux reconfiguration-\n");
            unlink(restartfile);
            return 0;
         }
      }

      while ( nread < 5 && sxread(tty_aux, buff + nread, 1, 1) ) {
         nread++;
      }

      if ( nread < 5 ) {
         is_idle = 1;
         nread = 0;
         count = 0;
         in_a_row = 0;
         max_in_a_row = config.aux_repcounts[MaxCount];
         if ( rfflood > 0 )
            send_rf_flood_message(0);
         rfflood = 0;
         continue;
      }
      is_idle = 0;

      /* Check for a continuous flood of RF inputs without a break */
      if ( ++in_a_row == max_in_a_row ) {
         rfflood = 1;
         send_rf_flood_message(SEC_FLOOD);
         max_in_a_row *= config.aux_repcounts[FloodRep];
         if ( max_in_a_row > ABSMAX_IN_A_ROW ) {
            max_in_a_row = ABSMAX_IN_A_ROW;
            in_a_row = 0;
         }
         continue;
      }

      /* Check for correct frame characters */
      if ( buff[0] != 0xD5u || buff[1] != 0xAAu || buff[4] != 0xADu ) {
         /* Alignment error - try to realign */
         type = RF_NOISEMR26;
         if ( config.disp_raw_rf & DISPMODE_RF_NOISE && rfflood == 0 )
            send_virtual_aux_data(buff[0], buff[1], RF_NOISEMR26, buff[2], buff[3], buff[4]);
         nread--;
         for ( k = 1; k < 5; k++ ) {
            if ( buff[k] == 0xD5u ) {
               memmove(buff, buff + k, nread);
               break;
            }
            nread--;
         }
         continue;
      }

      /* X10 Standard RF (type 0) or Entertainment RF (type 1) */
      type = (buff[2] & 0x02) ? RF_ENT : RF_STD;
      this_word = (buff[2] << 8 | buff[3]);

      /* Could use a more sophisticated validation step here */
      if ( type == RF_STD && (this_word & NOTX10_MASK) ) {
         if ( config.disp_raw_rf & DISPMODE_RF_NOISE && rfflood == 0 )
            send_virtual_aux_data(buff[0], buff[1], RF_NOISEMR26, buff[2], buff[3], buff[4]);
         nread = 0;
         continue;
      }

      if ( (config.disp_raw_rf & DISPMODE_RF_NORMAL) && !new_burst && rfflood == 0 ) {
         send_virtual_aux_data(buff[0], buff[1], RF_RAWMR26, buff[2], buff[3], buff[4]);
      }
      new_burst = 0;

      if ( count == 0 ) {
         rfword = this_word;
         count++;
         nread = 0;
         if ( type == RF_STD ) {
            rf_parms((int)rfword, &hcode, &ucode, &fcode);
            if ( count == config.aux_repcounts[MinCount] ) {
               proc_type_std(hcode, ucode, fcode);
            }
         }
         else  {
            if ( count == config.aux_repcounts[MinCount] ) {
               send_virtual_aux_data(0, buff[3], type, buff[2], 0, 0);
            }
         }
      }
      else if ( this_word == rfword ) {
         /* Repeat of previous burst */
         count++;
         nread = 0;
         if ( config.aux_repcounts[RepCount] > config.aux_repcounts[MinCount] &&
              (count % config.aux_repcounts[RepCount]) == config.aux_repcounts[MinCount] &&
               count < config.aux_repcounts[MaxCount] ) {
            if ( type == RF_STD ) {
               rf_parms((int)rfword, &hcode, &ucode, &fcode);
               proc_type_std(hcode, ucode, fcode);
            }
            else {
               send_virtual_aux_data(0, buff[3], type, buff[2], 0, 0);
            }
         }
      }
      else {
         /* New burst, save buffer */
         new_burst = 1;
         count = 0;
      }
   }
   return 0;
}



